"""
Manages the state of the axibot.

Possible system states:

    idle, ready to plot         (idle)
    plotting document           (plotting)
    canceling plotting          (canceling)
    processing document         (processing)

Possible state transitions are:

    idle -> plotting            (begin plotting)
    idle -> processing          (upload)
    processing -> idle          (done)
    plotting -> canceling       (cancel requested)
    canceling -> idle           (canceled)

Additional state variables are:

    active document (SVG text)
    list of actions grouped by path
    current path index
    current progress?

The app also holds a reference to the active EiBotBoard instance.
"""
from enum import Enum


class State(Enum):
    idle = 1
    plotting = 2
    canceling = 3
    processing = 4



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
MAX_RETRIES = 100

# These are unitless timing values used by the EBB.
SERVO_MIN = 7500
SERVO_MAX = 28000

# Default
SERVO_SPEED = 150

# Milliseconds?
EXTRA_PEN_UP_DELAY = 0
EXTRA_PEN_DOWN_DELAY = 0

SPEED_SCALE = 24.950  # ticks per millisecond

# Milliseconds of acceleration to reach full speed with pen down.
ACCEL_TIME_PEN_DOWN = 250

# Speed in ticks per millisecond.
SPEED_PEN_DOWN = 0.25 * SPEED_SCALE

# Milliseconds of acceleration to reach full speed with pen up.
ACCEL_TIME_PEN_UP = 1000

# Speed in ticks per millisecond.
SPEED_PEN_UP = 0.75 * SPEED_SCALE

# Short-move pen-up distance threshold in inches, below which we use the faster
# pen-down acceleration rate.
SHORT_THRESHOLD = 1.0

# Motor steps per inch in 16X microstepping mode.
DPI_16X = 2032

# Time interval in milliseconds to update motor control.
TIME_SLICE = 30

# Smoothness of curves. Units are inches.
CURVE_RESOLUTION = 0.02

# Skip pen-up moves shorter than this distance when possible. Units in inches.
MIN_GAP = 0.010

# Positions in arbitrary servo range units, from 0 to 100.
PEN_UP_POSITION = 60
PEN_DOWN_POSITION = 50



import logging

import aiohttp
from aiohttp import web

from . import api, plotting
from .state import State

log = logging.getLogger(__name__)


def broadcast(app, msg, exclude_client=None):
    s = msg.serialize()
    for ws in app['clients']:
        if ws != exclude_client:
            ws.send_str(s)


def notify_state(app, specific_client=None, exclude_client=None):
    state = app['state']
    num_actions = len(app['job'])
    action_index = app['action_index']
    msg = api.StateMessage(
        state=state.name,
        num_actions=num_actions,
        action_index=action_index,
        estimated_time=app['estimated_time'],
        consumed_time=app['consumed_time'],
        x=app['position'][0],
        y=app['position'][1],
        pen_up=app['pen_up'],
    )
    if specific_client:
        specific_client.send_str(msg.serialize())
    else:
        broadcast(app, msg, exclude_client=exclude_client)


def notify_new_document(app, exclude_client=None):
    msg = api.NewDocumentMessage(document=app['job'].document,
                                 filename=app['job'].filename)
    broadcast(app, msg, exclude_client=exclude_client)


def notify_error(app, to_client, s):
    msg = api.ErrorMessage(s)
    to_client.send_str(msg.serialize())


def notify_job_complete(app):
    log.info("Notifying clients of completed job. Est: %s, Actual: %s.",
             app['estimated_time'], app['consumed_time'])
    msg = api.CompletedJobMessage(estimated_time=app['estimated_time'],
                                  actual_time=app['consumed_time'])
    broadcast(app, msg)


async def handle_user_message(app, ws, msg):
    if isinstance(msg, api.SetDocumentMessage):
        assert app['state'] == State.idle
        try:
            app['state'] = State.processing
            notify_state(app)
            job = await plotting.process_upload_background(app,
                                                           msg.document,
                                                           msg.filename)
        except Exception as e:
            notify_error(app, ws, str(e))
        else:
            app['job'] = job
            app['estimated_time'] = job.duration().total_seconds()
            notify_new_document(app)
        finally:
            app['state'] = State.idle
            notify_state(app)

    elif isinstance(msg, api.ManualPenUpMessage):
        assert app['state'] == State.idle
        plotting.manual_pen_up(app)

    elif isinstance(msg, api.ManualPenDownMessage):
        assert app['state'] == State.idle
        plotting.manual_pen_down(app)

    elif isinstance(msg, api.ResumePlottingMessage):
        assert app['state'] == State.idle
        plotting.resume(app)
        notify_state(app)

    elif isinstance(msg, api.CancelPlottingMessage):
        assert app['state'] == State.plotting
        plotting.cancel(app)
        notify_state(app)

    else:
        log.error("Unknown user message: %s, ignoring.", msg)


async def client_handler(request):
    ws = web.WebSocketResponse()
    await ws.prepare(request)
    app = request.app

    log.info("Client connected.")
    clients = app['clients']
    clients.add(ws)

    notify_state(app, specific_client=ws)

    try:
        async for raw_msg in ws:
            if raw_msg.tp == aiohttp.MsgType.text:
                msg = api.Message.deserialize(raw_msg.data)
                log.info("User message: %s", msg)
                await handle_user_message(app, ws, msg)
            elif raw_msg.tp == aiohttp.MsgType.closed:
                break
            elif raw_msg.tp == aiohttp.MsgType.error:
                log.info("User websocket error: %s", raw_msg)
                break
            else:
                log.error("Unknown user message type: %s, ignoring.",
                          raw_msg.tp)
    finally:
        log.info("Client connection closed.")
        clients.remove(ws)

    return ws



import math

from .. import planning, config


vmax = config.SPEED_PEN_DOWN
accel_max = vmax / config.ACCEL_TIME_PEN_DOWN
timeslice = config.TIME_SLICE


def find_peak_speed(dtarray):
    vpeak = xlast = 0
    for x, duration in dtarray:
        xdiff = x - xlast
        v = xdiff / duration
        if v > vpeak:
            vpeak = v
        xlast = x

    # Fudge factor for float error
    vpeak -= 0.00001

    return vpeak


def test_trapezoidal_from_rest():
    dist = 9000
    dtarray = planning.interpolate_distance(dist, 0, 0,
                                            vmax, accel_max, timeslice)
    total_dist = dtarray[-1][0]
    assert dist == total_dist
    assert find_peak_speed(dtarray) <= vmax


def test_triangular_from_rest():
    dist = 3780
    dtarray = planning.interpolate_distance(dist, 0, 0,
                                            vmax, accel_max, timeslice)
    total_dist = dtarray[-1][0]
    assert dist == total_dist
    assert find_peak_speed(dtarray) <= vmax


def test_linear_changing_speed():
    dist = 150
    dtarray = planning.interpolate_distance(dist, 0.7, 0.75,
                                            vmax, accel_max, timeslice)
    total_dist = dtarray[-1][0]
    assert dist == total_dist
    assert find_peak_speed(dtarray) <= vmax


def test_cornering_angle_straight():
    angle = planning.cornering_angle((0, 0), (1, 1), (2, 2))
    assert angle == math.pi


def test_cornering_angle_reverse():
    angle = planning.cornering_angle((0, 0), (1, 1), (0, 0))
    assert angle == 0


def test_cornering_angle_90():
    angle = planning.cornering_angle((0, 0), (0, 5), (5, 5))
    assert angle == math.pi / 2.0


def test_cornering_speed_straight():
    vmax = config.SPEED_PEN_DOWN
    vstraight = planning.cornering_speed(math.pi, vmax)
    assert vmax == vstraight


def test_cornering_speed_reverse():
    vmax = config.SPEED_PEN_DOWN
    v = planning.cornering_speed(0, vmax)
    assert v == 0



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from datetime import timedelta
import json

from .action import PenUpMove, PenDownMove, XYMove, XYAccelMove, ABMove


class Job(list):
    def __init__(self, *args, pen_up_position, pen_down_position, servo_speed,
                 document=None, filename=None):
        self.filename = filename
        self.document = document
        self.pen_up_position = pen_up_position
        self.pen_down_position = pen_down_position
        self.servo_speed = servo_speed
        list.__init__(self, *args)

    def duration(self):
        duration_ms = sum(action.time() for action in self)
        return timedelta(seconds=(duration_ms / 1000))

    def serialize(self, f):
        actions = []
        for action in self:
            d = action.__dict__
            d['name'] = action.name
            actions.append(d)
        obj = {
            'filename': self.filename,
            'document': self.document,
            'pen_up_position': self.pen_up_position,
            'pen_down_position': self.pen_down_position,
            'servo_speed': self.servo_speed,
            'actions': actions,
        }
        json.dump(obj, f, indent=2)

    @classmethod
    def deserialize(cls, f):
        classes = (PenUpMove, PenDownMove, XYMove, XYAccelMove, ABMove)
        registry = {action_class.name: action_class
                    for action_class in classes}
        obj = json.load(f)
        actions = []
        for action_dict in obj.pop('actions'):
            name = action_dict.pop('name')
            action_class = registry[name]
            action = action_class(**action_dict)
            actions.append(action)
        return cls(actions, **obj)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import re
import math

from svg.path import Line, Arc, QuadraticBezier, CubicBezier


identity = [[1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0]]


def _flatten(mat):
    """
    Convert array format:
        [[A, C, E],
         [B, D, F]]
         ->
        [A, B, C, D, E, F]
    """
    return [mat[0][0], mat[1][0], mat[0][1], mat[1][1], mat[0][2], mat[1][2]]


def parse(s, base=identity):
    """
    Parse SVG transform syntax into a 2x3 affine transform matrix.
    """
    s = (s or "").strip()
    if not s:
        return base

    m = re.match(
        "(translate|scale|rotate|skewX|skewY|matrix)\s*\(([^)]*)\)\s*,?",
        s)

    cmd = m.group(1)
    val = m.group(2)

    if cmd == "translate":
        args = val.replace(',', ' ').split()
        dx = float(args[0])
        if len(args) == 1:
            dy = 0.0
        else:
            dy = float(args[1])
        matrix = [[1, 0, dx], [0, 1, dy]]

    elif cmd == "scale":
        args = val.replace(',', ' ').split()
        sx = float(args[0])
        if len(args) == 1:
            sy = sx
        else:
            sy = float(args[1])
        matrix = [[sx, 0, 0], [0, sy, 0]]

    elif cmd == "rotate":
        args = val.replace(',', ' ').split()
        a = float(args[0]) * math.pi / 180
        if len(args) == 1:
            cx, cy = (0.0, 0.0)
        else:
            cx, cy = map(float, args[1:])
        matrix = [[math.cos(a), -math.sin(a), cx],
                  [math.sin(a), math.cos(a), cy]]
        matrix = compose(matrix, [[1, 0, -cx], [0, 1, -cy]])

    elif cmd == "skewX":
        a = float(val) * math.pi / 180
        matrix = [[1, math.tan(a), 0], [0, 1, 0]]

    elif cmd == "skewY":
        a = float(val) * math.pi / 180
        matrix = [[1, 0, 0], [math.tan(a), 1, 0]]

    elif cmd == "matrix":
        a11, a21, a12, a22, v1, v2 = val.replace(',', ' ').split()
        matrix = [[float(a11), float(a12), float(v1)],
                  [float(a21), float(a22), float(v2)]]

    matrix = compose(base, matrix)
    if m.end() < len(s):
        return parse(s[m.end():], matrix)
    else:
        return matrix


def compose(mbase, mnew):
    """
    Compose a new affine transform matrix based on an existing one and a new
    one.
    """
    a11 = mbase[0][0] * mnew[0][0] + mbase[0][1] * mnew[1][0]
    a12 = mbase[0][0] * mnew[0][1] + mbase[0][1] * mnew[1][1]
    a21 = mbase[1][0] * mnew[0][0] + mbase[1][1] * mnew[1][0]
    a22 = mbase[1][0] * mnew[0][1] + mbase[1][1] * mnew[1][1]

    v1 = mbase[0][0] * mnew[0][2] + mbase[0][1] * mnew[1][2] + mbase[0][2]
    v2 = mbase[1][0] * mnew[0][2] + mbase[1][1] * mnew[1][2] + mbase[1][2]
    return [[a11, a12, v1], [a21, a22, v2]]


def apply_to_point(pt, matrix):
    """
    Apply a 2x3 affine transform matrix to a point.
    """
    x = matrix[0][0] * pt.real + matrix[0][1] * pt.imag + matrix[0][2]
    y = matrix[1][0] * pt.real + matrix[1][1] * pt.imag + matrix[1][2]
    return complex(x, y)


def apply_to_ellipse(rx, ry, ax, m):
    """
    Given an ellipse centered at 0, 0 defined by rx (x-radius), ry (y-radius),
    ax (angle x-axis is rotated), apply a flattened transform matrix. Then
    return new rx, ry, ax.
    """
    epsilon = 0.0000000001

    c = math.cos(math.radians(ax))
    s = math.sin(math.radians(ax))

    ma = [rx * (m[0] * c + m[2] * s),
          rx * (m[1] * c + m[3] * s),
          ry * (-m[0] * s + m[2] * c),
          ry * (-m[1] * s + m[3] * c)]

    j = ma[0] * ma[0] + ma[2] * ma[2]
    k = ma[1] * ma[1] + ma[3] * ma[3]

    d = (((ma[0] - ma[3]) * (ma[0] - ma[3]) +
          (ma[2] + ma[1]) * (ma[2] + ma[1])) *
         ((ma[0] + ma[3]) * (ma[0] + ma[3]) +
          (ma[2] - ma[1]) * (ma[2] - ma[1])))

    jk = (j + k) / 2

    if (d < epsilon * jk):
        new_rx = new_ry = math.sqrt(jk)
        new_ax = 0
        return new_rx, new_ry, new_ax

    l = ma[0] * ma[1] + ma[2] * ma[3]

    d = math.sqrt(d)

    l1 = jk + d / 2
    l2 = jk - d / 2

    if abs(l) < epsilon and abs(l1 - k) < epsilon:
        new_ax = 90
    else:
        if abs(l) > abs(l1 - k):
            new_ax = math.degrees(math.atan((l1 - j) / l))
        else:
            new_ax = math.degrees(math.atan(l / (l1 - k)))

    if ax >= 0:
        new_rx = math.sqrt(l1)
        new_ry = math.sqrt(l2)
    else:
        new_ax += 90
        new_rx = math.sqrt(l2)
        new_ry = math.sqrt(l1)

    return new_rx, new_ry, new_ax


def apply(path, matrix):
    """
    Apply an affine transform to a Path instance.

    XXX this is broken for arcs
    """
    for piece in path:
        if isinstance(piece, Line):
            piece.start = apply_to_point(piece.start, matrix)
            piece.end = apply_to_point(piece.end, matrix)
        elif isinstance(piece, Arc):
            ma = _flatten(matrix)
            rx, ry, ax = apply_to_ellipse(
                piece.radius.real,
                piece.radius.imag,
                piece.rotation,
                ma)
            if ma[0] * ma[3] - ma[1] * ma[2] < 0:
                piece.sweep = not piece.sweep
            piece.start = apply_to_point(piece.start, matrix)
            piece.end = apply_to_point(piece.end, matrix)
            piece.center = apply_to_point(piece.center, matrix)
            piece.radius = complex(rx, ry)
            piece.rotation = ax
        elif isinstance(piece, QuadraticBezier):
            piece.start = apply_to_point(piece.start, matrix)
            piece.end = apply_to_point(piece.end, matrix)
            piece.control = apply_to_point(piece.control, matrix)
        elif isinstance(piece, CubicBezier):
            piece.start = apply_to_point(piece.start, matrix)
            piece.end = apply_to_point(piece.end, matrix)
            piece.control1 = apply_to_point(piece.control1, matrix)
            piece.control2 = apply_to_point(piece.control2, matrix)
        else:
            raise ValueError("Don't know how to transform %r" % piece)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import os.path

from .. import svg, config

__here__ = os.path.dirname(__file__)
example_dir = os.path.join(__here__, '..', '..', 'examples')


def test_smoke():
    filename = os.path.join(example_dir, 'mixed.svg')
    with open(filename) as f:
        paths = svg.extract_paths(f.read())
    assert paths

    segments = svg.plan_segments(paths, resolution=config.CURVE_RESOLUTION)
    assert segments

    segments = svg.add_pen_up_moves(segments)
    assert segments



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from ..job import Job
from ..action import XYMove, PenUpMove, PenDownMove


def test_roundtrip():
    job = Job(pen_up_position=60, pen_down_position=40, servo_speed=150)
    job.append(PenDownMove(400))
    job.append(XYMove(500, 300, 200))
    job.append(PenUpMove(400))

    testfile = '/tmp/test.axibot.json'

    with open(testfile, 'w') as f:
        job.serialize(f)

    with open(testfile, 'r') as f:
        newjob = Job.deserialize(f)

    assert job == newjob



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os.path
from axibot.cmd import main

from . import utils


def test_info_smoke():
    filename = os.path.join(utils.example_dir, 'circles.svg')
    args = ['axibot', 'info', filename]
    main(args)



import logging

import math
from six.moves import StringIO

from .. import planning, config
from ..job import Job
from ..action import PenUpMove, PenDownMove, XYMove

from . import handlers
from .state import State

log = logging.getLogger(__name__)


def estimate_time(actions):
    return sum(action.time() for action in actions) / 1000.


def step_segments_to_actions(app, step_segments):
    pen_up_delay = app['pen_up_delay']
    pen_down_delay = app['pen_down_delay']
    segments_limits = planning.plan_speed(step_segments)
    return planning.plan_actions(segments_limits,
                                 pen_up_delay=pen_up_delay,
                                 pen_down_delay=pen_down_delay)


def plan_pen_up_move(app, start, end):
    step_segments = [((start, end), True)]
    actions = step_segments_to_actions(app, step_segments)
    return actions


def plan_deceleration(app, position, v):
    """
    Given an initial position in steps, and an initial velocity vector in steps
    per millisecond, create a series of actions which will decelerate as
    quickly as possible, moving in the same direction.

    Return a tuple of (end point, actions)
    """
    if app['pen_up']:
        vmax = config.SPEED_PEN_UP
        accel_time = config.ACCEL_TIME_PEN_UP
    else:
        vmax = config.SPEED_PEN_DOWN
        accel_time = config.ACCEL_TIME_PEN_DOWN

    # Calculate distance required for deceleration.
    vx, vy = v
    vmag = math.sqrt((vx**2) + (vy**2))
    x_accel_time = (abs(vx) / vmax) * accel_time
    x_accel_dist = x_accel_time * (vx / 2)
    y_accel_time = (abs(vy) / vmax) * accel_time
    y_accel_dist = y_accel_time * (vy / 2)

    # Ugh
    if vmag > vmax:
        vmag = vmax

    # Make a line extending from this position along that distance.
    end = (int(round(position[0] + x_accel_dist)),
           int(round(position[1] + y_accel_dist)))

    # Assert that it doesn't exceed the bounds of the machine.
    # XXX

    if position == end:
        # Don't need to decelerate.
        return position, []

    # Plan the actions for this deceleration segment.
    dtarray = planning.interpolate_pair(position, vmag,
                                        end, 0, app['pen_up'])
    return end, planning.dtarray_to_moves(position, end, dtarray)


def process_upload(app, document, filename):
    if document[0] == '{':
        f = StringIO(document)
        return Job.deserialize(f)
    else:
        return planning.plan_job(document, filename=filename)


async def process_upload_background(app, document, filename):
    return await app.loop.run_in_executor(None, process_upload, app,
                                          document, filename)


def update_bot_state(app, action):
    if isinstance(action, XYMove):
        dx = (action.m1 + action.m2) / 2
        dy = (action.m1 - action.m2) / 2
        lastx, lasty = app['position']
        app['position'] = lastx + dx, lasty + dy
        app['consumed_time'] += (action.duration / 1000.)
    elif isinstance(action, PenUpMove):
        app['pen_up'] = True
        app['consumed_time'] += (action.delay / 1000.)
    elif isinstance(action, PenDownMove):
        app['pen_up'] = False
        app['consumed_time'] += (action.delay / 1000.)


async def cancel_to_origin(app, action):
    if isinstance(action, PenUpMove):
        v = 0, 0
    elif isinstance(action, PenDownMove):
        v = 0, 0
    elif isinstance(action, XYMove):
        dx = (action.m1 + action.m2) / 2
        dy = (action.m1 - action.m2) / 2
        v = dx / action.duration, dy / action.duration
    else:
        raise ValueError("don't understand action: %r" % action)

    # plan trajectory from this v to zero
    decel_end, actions = plan_deceleration(app, app['position'], v)

    # lift pen up if not already up
    if not app['pen_up']:
        actions.append(PenUpMove(app['pen_up_delay']))

    # plan move back to origin
    actions.extend(plan_pen_up_move(app, decel_end, (0, 0)))
    orig_estimated = app['estimated_time']
    app['estimated_time'] = estimate_time(actions)
    app['consumed_time'] = 0
    bot = app['bot']

    for action in actions:
        def run_action():
            bot.do(action)
        update_bot_state(app, action)
        await app.loop.run_in_executor(None, run_action)
        handlers.notify_state(app)

    app['estimated_time'] = orig_estimated
    app['consumed_time'] = 0


async def plot_task(app):
    log.debug("plot_task: begin")
    app['state'] = State.plotting
    bot = app['bot']
    app['consumed_time'] = 0

    if app['pen_up'] is not True:
        bot.pen_up(app['pen_up_delay'])
        app['pen_up'] = True

    while True:
        job = app['job']
        action_index = app['action_index']
        action = job[action_index]

        def run_action():
            bot.do(action)

        update_bot_state(app, action)
        await app.loop.run_in_executor(None, run_action)
        action_index += 1
        if action_index == len(job):
            # Finished
            log.debug("plot_task: plotting complete")
            handlers.notify_job_complete(app)
            app['consumed_time'] = 0
            break

        app['action_index'] = action_index

        if app['state'] == State.canceling:
            log.debug("plot_task: canceling")
            # Decelerate, pen up, fastest move back to origin
            await cancel_to_origin(app, action)
            break
        # notify clients of state change
        handlers.notify_state(app)

    app['state'] = State.idle
    app['action_index'] = 0

    # send job complete message
    # notify clients of state change
    handlers.notify_state(app)
    log.warn("plot_task: end")


async def manual_task(app, action):
    orig_state = app['state']
    log.debug("manual task: set state to plotting")
    app['state'] = State.plotting
    handlers.notify_state(app)
    bot = app['bot']

    def run():
        bot.do(action)

    await app.loop.run_in_executor(None, run)
    app['state'] = orig_state
    log.debug("manual task: returned state to %s", orig_state)
    handlers.notify_state(app)


def manual_pen_up(app):
    pen_up_delay = app['pen_up_delay']
    app.loop.create_task(manual_task(app, PenUpMove(pen_up_delay)))


def manual_pen_down(app):
    pen_down_delay = app['pen_down_delay']
    app.loop.create_task(manual_task(app, PenDownMove(pen_down_delay)))


def resume(app):
    app.loop.create_task(plot_task(app))


def cancel(app):
    app['state'] = State.canceling
    handlers.notify_state(app)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging

import math

from . import config, svg
from .action import XYMove, PenUpMove, PenDownMove
from .job import Job

log = logging.getLogger(__name__)


def calculate_pen_delays(up_position, down_position, servo_speed):
    """
    The AxiDraw motion controller must know how long to wait after giving a
    'pen up' or 'pen down' command. This requires calculating the speed that
    the servo can move to or from the two respective states. This function
    performs that calculation and returns a tuple of (pen_up_delay,
    pen_down_delay). All delays are in milliseconds.
    """
    assert up_position > down_position

    # Math initially taken from axidraw inkscape driver, but I think this can
    # be sped up a bit. We might also want to use different speeds for up/down,
    # due to the added weight of the pen slowing down the servo in the 'up'
    # direction.
    dist = up_position - down_position
    time = int((1000. * dist) / config.SERVO_SPEED)

    return ((time + config.EXTRA_PEN_UP_DELAY),
            (time + config.EXTRA_PEN_DOWN_DELAY))


def distance(a, b):
    """
    Distance between two points.
    """
    return math.sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2)


def convert_inches_to_steps(segments):
    """
    Take the output from add_pen_up_moves() and convert all points from inches
    to steps.

    This also 'collapses points': that is, if there are two or more adjacent
    points in a segment are at the same position, they are combined into one.
    """
    spi = config.DPI_16X
    out = []
    for segment, pen_up in segments:
        points = []
        last_point = None
        for inches_point in segment:
            assert len(inches_point) == 2
            steps_point = (int(round(spi * inches_point[0])),
                           int(round(spi * inches_point[1])))
            if (last_point is None) or (steps_point != last_point):
                points.append(steps_point)
            last_point = steps_point
        out.append((points, pen_up))
    return out


def cornering_angle(a, b, c):
    """
    Given three points, compute the angle in radians between AB and BC.
    """
    assert a != b
    assert b != c

    ax, ay = a
    bx, by = b
    cx, cy = c
    x = (bx - ax)**2 + (by - ay)**2
    y = (bx - cx)**2 + (by - cy)**2
    z = (cx - ax)**2 + (cy - ay)**2

    arg = (x + y - z) / math.sqrt(4 * x * y)

    # Floating point error can accumulate here and give us an argument to
    # acos() that will be like 1.00000000000whatever
    # Catch that and just make it 1 or -1.

    if arg > 1:
        assert arg < 1.000002
        arg = 1
    elif arg < -1:
        assert arg > -1.000002
        arg = -1

    return math.acos(arg)


def cornering_speed(angle, vmax):
    """
    Given a corner angle in radians, compute the desired cornering speed.

    An angle of pi can go full speed. An angle of zero needs to fully changeh
    direction, so it needs to have zero speed. Everywhere in between...?

    0 < angle < pi/2 should be zero speed
    pi / 2 < angle < pi should be linear interpolation?

    XXX figure this out
    """
    assert angle <= math.pi
    if angle < (math.pi / 2):
        return 0
    else:
        return (1.0 + math.sin(angle - math.pi)) * vmax


def segment_corner_limits(segment, pen_up):
    """
    Given a segment and pen state, tag each point with the 'speed limit' for
    that corner. Ensure that the segment starts and ends at zero speed.
    """
    out = []
    out.append((segment[0], 0.0))

    vmax = config.SPEED_PEN_UP if pen_up else config.SPEED_PEN_DOWN

    for a, b, c in zip(segment[:-2], segment[1:-1], segment[2:]):
        angle = cornering_angle(a, b, c)
        limit = cornering_speed(angle, vmax)
        out.append((b, limit))

    out.append((segment[-1], 0.0))
    return out


def segment_acceleration_limits(segment, pen_up):
    """
    Given a segment w/ speed limits and pen state, tag each point with the
    target speed for that corner. This will possibly reduce speeds from the
    speed limit, to account for acceleration and deceleration requirements.

    We'll do two passes through the list of points: one forward for
    acceleration, one backward for deceleration.
    """
    assert segment
    if pen_up:
        vmax = config.SPEED_PEN_UP
        accel_time = config.ACCEL_TIME_PEN_UP
    else:
        vmax = config.SPEED_PEN_DOWN
        accel_time = config.ACCEL_TIME_PEN_DOWN
    accel_rate = vmax / accel_time

    out = []
    last_point, last_speed = segment[0]
    out.append((last_point, last_speed))
    for point, speed_limit in segment[1:]:
        dist = distance(point, last_point)
        # can we accelerate from last_speed to speed_limit in this distance?
        top_speed = math.sqrt((2 * accel_rate * dist) + last_speed**2)
        # if not we need to limit the speed at this point
        speed = min(speed_limit, top_speed)
        assert speed <= vmax
        out.append((point, speed))
        last_point = point
        last_speed = speed

    assert out
    return out


def plan_speed(segments):
    """
    Given a list of (segment, pen_up) tuples, tag each segment with a target
    speed for that corner. This combines two limits:

    1. The max speed at which we can plot through a given corner, based on how
    tight that corner is and the pen state.
    2. The acceleration/deceleration limit.

    Both of these limits serve to avoid the pen carriage oscillating and thus
    making the pen line squiggly, or just slamming the motors around and
    causing wear on the machine.
    """
    out = []
    for segment, pen_up in segments:
        assert segment
        points = segment_corner_limits(segment, pen_up)
        # Calculate forward acceleration
        points = segment_acceleration_limits(points, pen_up)
        assert points
        # Calculate reverse acceleration (deceleration)
        points = segment_acceleration_limits(points[::-1], pen_up)
        assert points
        points = points[::-1]
        assert points
        out.append((points, pen_up))
    return out


def check_limits(dots):
    xend, yend, duration = dots[-1]
    return xend, yend


def mess_with_dots(start, end, dots):
    if not dots:
        return []
    x_desired = end[0] - start[0]
    y_desired = end[1] - start[1]
    x_got, y_got = check_limits(dots)

    if x_got:
        x_adjust = x_desired / x_got
    else:
        assert x_desired == 0
        x_adjust = 0
    if y_got:
        y_adjust = y_desired / y_got
    else:
        assert y_desired == 0
        y_adjust = 0

    new_dots = [(int(round(x * x_adjust)), int(round(y * y_adjust)), duration)
                for x, y, duration
                in dots]
    assert check_limits(new_dots) == (x_desired, y_desired)
    return new_dots


def dtarray_to_moves(start, end, dtarray):
    """
    Given a start point, an end point, and an array of linear distances/times,
    return the actions to move between the two points, with one move per point
    in the array.
    """
    assert end != start

    dist = distance(end, start)
    # assert end_dist == dist, "expected %r == %r" % (dtarray[-1][0], dist)
    xratio = (end[0] - start[0]) / dist
    yratio = (end[1] - start[1]) / dist

    dots = []
    for d, duration in dtarray:
        assert duration < 3000, "tried to set duration:%r" % duration
        x = int(round(xratio * d))
        y = int(round(yratio * d))
        duration = int(round(duration))
        dots.append((x, y, duration))

    dots = mess_with_dots(start, end, dots)

    actions = []
    prev_x, prev_y = 0, 0
    check_x, check_y = start
    for x, y, duration in dots:
        dx = x - prev_x
        dy = y - prev_y
        prev_x = x
        prev_y = y
        check_x += dx
        check_y += dy
        if dx or dy:
            # Convert to AxiDraw coordinate space.
            m1 = dx + dy
            m2 = dx - dy
            actions.append(XYMove(m1=m1, m2=m2, duration=duration))

    check = check_x, check_y
    assert check == end, \
        "actually ended at %r, wanted to end at %r" % (check, end)

    return actions


def interpolate_distance_trapezoidal(vstart, accel_time, accel_dist,
                                     vend, decel_time, decel_dist,
                                     vmax, dist, timeslice):
    accel_slices = int(math.floor(accel_time / timeslice))
    decel_slices = int(math.floor(decel_time / timeslice))

    dtarray = []

    x = 0
    v = vstart

    if accel_slices:
        accel_timeslice = accel_time / accel_slices
        assert accel_timeslice > timeslice, \
            "accel timeslice %r too small" % accel_timeslice
        vstep = (vmax - vstart) / (accel_slices + 1)
        for n in range(accel_slices):
            v += vstep
            if v > 0:
                x += v * accel_timeslice
                dtarray.append((x, accel_timeslice))

    # This could just use a single action to move for the constant-speed
    # section of the trapezoidal speed profile, but by slicing it up into
    # timeslices, we can make the web interface visualization and such
    # look smoother.
    coast_dist = dist - (accel_dist + decel_dist)
    coast_time = coast_dist / vmax
    coast_slices = int(math.floor(coast_time / timeslice))
    if coast_slices:
        coast_timeslice = coast_time / coast_slices
        assert coast_timeslice > timeslice, \
            "coast_timeslice %r too small" % coast_timeslice
        for n in range(coast_slices):
            v = vmax
            x += v * coast_timeslice
            dtarray.append((x, coast_timeslice))

    if decel_slices:
        decel_timeslice = decel_time / decel_slices
        assert decel_timeslice > timeslice, \
            "decel timeslice %r too small" % decel_timeslice
        vstep = (vend - vmax) / (decel_slices + 1)
        for n in range(decel_slices):
            v += vstep
            if v > 0:
                x += v * decel_timeslice
                dtarray.append((x, decel_timeslice))

    return dtarray


def interpolate_distance_linear(dist, vstart, vend, accel_rate, timeslice):
    # Linear
    lin_accel = ((vend**2 - vstart**2) / (2 * dist))
    lin_accel = min(lin_accel, accel_rate)
    lin_accel = max(lin_accel, -accel_rate)
    lin_time = (vend - vstart) / lin_accel

    vavg = (vend + vstart) / 2
    lin_time = vavg / dist

    x = 0
    v = vstart
    dtarray = []

    slices = int(math.floor(lin_time / timeslice))
    if not slices:
        duration = dist / vavg
        # XXX this should be set more intelligently
        if duration > 200:
            duration = 200
        if duration < 30:
            duration = 30
        return [(dist, duration)]
    if slices:
        lin_timeslice = lin_time / slices
        assert lin_timeslice > timeslice, \
            "lin timeslice %r too small" % lin_timeslice
        vstep = (vend - vstart) / slices
        for n in range(slices):
            v += vstep
            if v > 0:
                x += v * lin_timeslice
                dtarray.append((x, lin_timeslice))
    else:
        # XXX ???
        dtarray.append((dist, timeslice))
    return dtarray


def interpolate_distance_triangular(dist, vstart, vend, accel_rate, timeslice):
    accel_time = ((math.sqrt((2 * vstart**2) +
                             (2 * vend**2) +
                             (4 * accel_rate * dist)) -
                   (2 * vstart)) /
                  (2 * accel_rate))
    accel_slices = int(math.floor(accel_time / timeslice))
    if accel_slices <= 0:
        accel_slices = 0
        accel_time = 0

    vmax = vstart + (accel_time * accel_rate)

    decel_time = accel_time - (vend - vstart) / accel_rate
    decel_slices = int(math.floor(decel_time / timeslice))

    dtarray = []
    x = 0
    v = vstart

    if (accel_slices + decel_slices) > 4:
        # Triangular
        if accel_slices:
            accel_timeslice = accel_time / accel_slices
            assert accel_timeslice > timeslice, \
                "accel timeslice %r too small" % accel_timeslice
            vstep = (vmax - vstart) / (accel_slices + 1.0)
            for n in range(accel_slices):
                v += vstep
                if v > 0:
                    x += v * accel_timeslice
                    dtarray.append((x, accel_timeslice))

        if decel_slices:
            decel_timeslice = decel_time / decel_slices
            assert decel_timeslice > timeslice, \
                "decel timeslice %r too small" % decel_timeslice
            vstep = (vend - vmax) / (decel_slices + 1.0)
            for n in range(decel_slices):
                v += vstep
                if v > 0:
                    x += v * decel_timeslice
                    dtarray.append((x, decel_timeslice))
    elif vend == vstart:
        if vstart:
            # Constant speed that is non-zero
            duration = dist / vstart
            # XXX this should be set more intelligently
            if duration > 200:
                duration = 200
            if duration < 30:
                duration = 30
            return [(dist, duration)]
        else:
            # Segment that has to start and end at zero speed, but is really
            # short. This is a really obnoxious case, for now we're just going
            # to set it to do the move in 100ms. Need to figure out the 'vmin'
            # which can be accelerated to instantaneously.
            return [(dist, 100)]
    else:
        dtarray = interpolate_distance_linear(dist, vstart, vend, accel_rate,
                                              timeslice)

    # assert end_dist == dist, "%r must == %r" % (dtarray[-1][0], dist)
    return dtarray


def interpolate_distance(dist, vstart, vend, vmax, accel_max, timeslice):
    """
    Given a distance to traverse, start and end velocities in the direction of
    movement, a maximum speed, an acceleration rate, and a minimum
    timeslice, generated an array of (distance, time) points.

    Distancec units are motor steps.
    Velocity units are motor steps per second.
    Acceleration units are steps/second^2.
    Timeslice provided is in seconds.
    """
    accel_time = (vmax - vstart) / accel_max
    decel_time = (vmax - vend) / accel_max
    accel_dist = (vstart * accel_time) + (0.5 * accel_max * (accel_time**2))
    decel_dist = (vend * decel_time) + (0.5 * accel_max * (decel_time**2))

    if dist > (accel_dist + decel_dist + timeslice * vmax):
        # Trapezoidal
        return interpolate_distance_trapezoidal(
            vstart, accel_time, accel_dist,
            vend, decel_time, decel_dist,
            vmax, dist, timeslice)
    else:
        # Triangular or linear
        return interpolate_distance_triangular(dist, vstart, vend, accel_max,
                                               timeslice)


def interpolate_pair(start, vstart, end, vend, pen_up):
    """
    Given start/end positions, velocities, and pen state, return the array of
    distance/time to move between two points.

    Note that a given distance entry is the distance travelled *after* the
    point/move has been made.

    We want to always be accelerating at a constant rate, decelerating at a
    constant rate, or moving at the maximum speed for this pen state.
    """
    if pen_up:
        vmax = config.SPEED_PEN_UP
        accel_max = vmax / config.ACCEL_TIME_PEN_UP
    else:
        vmax = config.SPEED_PEN_DOWN
        accel_max = vmax / config.ACCEL_TIME_PEN_DOWN

    timeslice = config.TIME_SLICE

    dist = distance(end, start)

    assert vstart <= vmax, "%f must be <= %f" % (vstart, vmax)
    assert vend <= vmax, "%f must be <= %f" % (vend, vmax)
    return interpolate_distance(dist, vstart, vend, vmax, accel_max, timeslice)


def interpolate_segment(segment, pen_up):
    """
    Given a segment with assigned speeds at each point, and a max
    acceleration/deceleration rate, create the moves to traverse the segment.
    """
    assert segment[0][1] == 0
    assert segment[-1][1] == 0

    actions = []
    # Iterate over point pairs.
    last_point, last_speed = segment[0]
    for point, speed in segment[1:]:
        if point != last_point:
            dist_array = interpolate_pair(last_point, last_speed,
                                          point, speed, pen_up)
            actions.extend(dtarray_to_moves(last_point, point, dist_array))
        last_point = point
        last_speed = speed
    return actions


def plan_actions(segments_with_speed, pen_up_delay, pen_down_delay):
    """
    Given a list of (segment, pen_up) tuples as returned by plan_speed(),
    return a list of moves. Also add any pen state changes.
    """
    actions = []
    last_pen_up = True
    for segment, pen_up in segments_with_speed:
        if pen_up != last_pen_up:
            if pen_up:
                actions.append(PenUpMove(pen_up_delay))
            else:
                actions.append(PenDownMove(pen_down_delay))
            last_pen_up = pen_up
        actions.extend(interpolate_segment(segment, pen_up))
    return actions


def plan_job(document, filename):
    pen_up_position = config.PEN_UP_POSITION
    pen_down_position = config.PEN_DOWN_POSITION
    servo_speed = config.SERVO_SPEED
    pen_up_delay, pen_down_delay = \
        calculate_pen_delays(pen_up_position, pen_down_position, servo_speed)

    log.info("Loading %s...", filename)
    log.info("Extracting paths...")
    paths = svg.extract_paths(document)
    paths = svg.preprocess_paths(paths)
    log.info("Planning segments...")
    segments = svg.plan_segments(paths, resolution=config.CURVE_RESOLUTION)
    log.info("Adding pen-up moves...")
    segments = svg.add_pen_up_moves(segments)
    log.info("Converting inches to steps...")
    step_segments = convert_inches_to_steps(segments)
    log.info("Planning speed limits...")
    segments_limits = plan_speed(step_segments)
    log.info("Planning actions...")
    actions = plan_actions(segments_limits,
                           pen_up_delay=pen_up_delay,
                           pen_down_delay=pen_down_delay)
    return Job(actions,
               pen_up_position=pen_up_position,
               pen_down_position=pen_down_position,
               servo_speed=servo_speed,
               document=document,
               filename=filename)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os.path
from axibot.debug import main

from . import utils


def test_debug_smoke():
    infile = os.path.join(utils.example_dir, 'rectangles.svg')
    outfile = '/tmp/axibot-test-debug.png'
    args = ['axibot-debug',
            'actions',
            '--out',
            outfile,
            infile]
    main(args)
    assert os.path.exists(outfile)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging

import math
import re

from xml.etree import ElementTree
from svg.path import parse_path, Path, Line

from . import transform

log = logging.getLogger(__name__)


def convert_to_inches(s):
    """
    Convert an SVG unit string like '42.5mm' or '12in' to inches as a float. If
    the unit string does not include physical units, return None.
    """
    assert s
    s = s.strip()
    for unit in ('in', 'mm', 'cm', 'px', 'pt'):
        if s[-2:] == unit:
            v = s[:-2]
            break
    else:
        return None
    v = float(v)
    if unit == 'in':
        return v
    elif unit == 'mm':
        return v / 25.4
    elif unit == 'cm':
        return v / 2.54
    elif unit == 'px':
        return v / 96.
    elif unit == 'pt':
        return v / 72.


def parse_pixels(s):
    # Both px and pt are effectively unitless?
    if s.endswith('px') or s.endswith('pt'):
        s = s[:-2]
    return float(s)


def get_viewbox_dimensions(viewbox):
    viewbox_fields = viewbox.strip().replace(',', ' ').split(' ')
    sx, sy, w, h = viewbox_fields
    return w, h


def get_document_dimensions(tree, max_area=(11.0, 8.5)):
    """
    Return the dimensions of this document in inches as to be plotted. If the
    document specifies physical units, they will be converted to inches, and
    asserted to be less than the working area of the AxiDraw. If the document
    does not specify physical units (e.g. the width and height are in pixels
    only) it will be scaled to the working area.

    Returns a tuple of (width, height) in inches.
    """
    max_width, max_height = max_area
    raw_width = tree.get('width')
    raw_height = tree.get('height')
    if not (raw_width and raw_height):
        log.warn("This document does not have width and height attributes. "
                 "Extracting viewbox dimensions.")
        svg_width = svg_height = None
        raw_width, raw_height = get_viewbox_dimensions(tree.get('viewBox'))
    else:
        svg_width = convert_to_inches(raw_width)
        svg_height = convert_to_inches(raw_height)
    if not (svg_width and svg_height):
        log.warn("This document does not specify physical units. "
                 "Auto-scaling it to fit the drawing area.")
        width = parse_pixels(raw_width)
        height = parse_pixels(raw_height)
        aspect_ratio = width / height
        max_ratio = max_width / max_height
        if aspect_ratio > max_ratio:
            # Wider than working area, constrained by width
            scale = max_width / width
        else:
            # Taller than working area, constrained by height
            scale = max_height / height
        svg_width = scale * width
        svg_height = scale * height
    assert svg_width <= max_width or math.isclose(svg_width, max_width), \
        "SVG width of %s must be <= %s" % (svg_width, max_width)
    assert svg_height <= max_height or math.isclose(svg_height, max_height), \
        "SVG height of %s must be <= %s" % (svg_height, max_height)
    return svg_width, svg_height


def svgns(tag):
    return '{http://www.w3.org/2000/svg}' + tag

def xlinkns(tag):
    return '{http://www.w3.org/1999/xlink}' + tag


def points_to_path_string(points_string, closed):
    points_string = points_string.replace(","," ")
    points = []
    split_points = points_string.split()
    for point_pair in zip(split_points[::2], split_points[1::2]):
        xs, ys = point_pair
        x = float(xs)
        y = float(ys)
        points.append((x, y))

    startx, starty = points[0]
    pieces = ['M{x},{y}'.format(x=startx, y=starty)]
    for x, y in points[1:]:
        pieces.append('L{x},{y}'.format(x=x, y=y))

    if closed:
        pieces.append('L{x},{y}'.format(x=startx, y=starty))

    return ' '.join(pieces)


def convert_to_path(node, matrix):
    if node.tag == svgns('rect'):
        # transform a rect element into a path
        x = float(node.get('x'))
        y = float(node.get('y'))
        w = float(node.get('width'))
        h = float(node.get('height'))
        return 'M{x},{y} l{w},0 l0,{h} l-{w},0 z'.format(x=x, y=y, w=w, h=h)

    elif node.tag == svgns('line'):
        # transform a line element into a path element
        x1 = float(node.get('x1'))
        y1 = float(node.get('y1'))
        x2 = float(node.get('x2'))
        y2 = float(node.get('y2'))
        return 'M{x1},{y1} L{x2},{y2}'.format(x1=x1, y1=y1, x2=x2, y2=y2)

    elif node.tag == svgns('polyline'):
        return points_to_path_string(node.get('points'), closed=False)
    elif node.tag == svgns('polygon'):
        return points_to_path_string(node.get('points'), closed=True)
    elif node.tag == svgns('ellipse') or node.tag == svgns('circle'):
        # Convert circles and ellipses to a path with two 180 degree arcs.
        # In general (an ellipse), we convert
        #   <ellipse rx="RX" ry="RY" cx="X" cy="Y"/>
        # to
        #   <path d="MX1,CY A RX,RY 0 1 0 X2,CY A RX,RY 0 1 0 X1,CY"/>
        # where
        #   X1 = CX - RX
        #   X2 = CX + RX
        if node.tag == svgns('ellipse'):
            rx = float(node.get('rx', '0'))
            ry = float(node.get('ry', '0'))
        else:
            rx = float(node.get('r', '0'))
            ry = rx
        # XXX handle rx or ry of zero?

        cx = float(node.get('cx', '0'))
        cy = float(node.get('cy', '0'))
        x1 = cx - rx
        x2 = cx + rx
        return ('M %f,%f ' % (x1, cy) +
                'A %f,%f ' % (rx, ry) +
                '0 1 0 %f,%f ' % (x2, cy) +
                'A %f,%f ' % (rx, ry) +
                '0 1 0 %f,%f' % (x1, cy))
    else:
        raise ValueError("Don't know how to convert %s tag to a path." %
                         node.tag)


def subdivide_path(path, resolution):
    """
    Given a svg.path.Path instance, output a list of points to traverse. The
    ``smoothness`` parameter specifies how smooth the curve approximation
    should be.

    Note that typical "full speed" distance per timeslice is around 0.37".
    """
    points = []
    for piece in path:
        if isinstance(piece, Line):
            # Don't subdivide lines.
            points.append((piece.start.real, piece.start.imag))
            points.append((piece.end.real, piece.end.imag))
        else:
            dist = piece.length(error=1e-6)
            count = int(math.ceil(dist / resolution))
            for n in range(count + 1):
                point = piece.point(n / count)
                points.append((point.real, point.imag))
    return points


def plan_segments(paths, resolution):
    """
    Takes a list of Path instances, returns a list of lists of points.
    """
    return [subdivide_path(path, resolution) for path in paths]


def join_segments(segments, min_gap):
    """
    Takes a list of segments (which are lists of points) and joins them when
    the start of one segment is within a certain tolerance of the end of the
    previous segment.

    XXX this needs to actually get called, maybe. Not sure it helps.
    """
    if len(segments) < 2:
        return segments
    last_segment = segments[0]
    new_segments = [last_segment]
    for segment in segments[1:]:
        x0, y0 = last_segment[-1]
        x1, y1 = segment[0]
        if math.sqrt((x1 - x0)**2 + (y1 - y0)**2) < min_gap:
            last_segment.extend(segment)
        else:
            last_segment = segment
            new_segments.append(segment)
    return new_segments


def add_pen_up_moves(segments):
    """
    Takes a list of pen-down segments. Returns a list of (segment, pen_up)
    tuples, with additional segments added that pen-up move between segments.
    Also add segments at the beginning and end to do pen-up moves from and to
    the origin location.

    The output list should thus always have 2n+1 elements, where n is the
    length of the input list.
    """
    assert segments
    origin = 0, 0

    out_segments = []
    start_seg = [origin, segments[0][0]]
    out_segments.append((start_seg, True))

    count = len(segments)

    for n, seg in enumerate(segments, start=1):
        assert seg
        out_segments.append((seg, False))
        if n == count:
            # last one
            next_seg_start = origin
        else:
            next_seg_start = segments[n][0]
        inter_seg = [seg[-1], next_seg_start]
        out_segments.append((inter_seg, True))

    return out_segments


def transform_path(s, transform_matrix):
    """
    Parse the path described by string ``s`` and apply the transform matrix.
    """
    p = parse_path(s)
    transform.apply(p, transform_matrix)
    return p


def recurse_tree(
    paths, tree_root, tree, transform_matrix, parent_visibility='visible'):
    """
    Append path tuples to the ``paths`` variable in place, while recursively
    parsing ``tree``.
    """
    for node in tree:
        log.debug("Handling element: %r: %r", node, node.items())

        # calculate node viz, ignore invisible nodes
        v = node.get('visibility', parent_visibility)
        if v == 'inherit':
            v = parent_visibility
        elif v == 'hidden' or v == 'collapse':
            pass

        # first apply current transform to this node's transform
        matrix_new = transform.compose(transform_matrix,
                                       transform.parse(node.get('transform')))

        if node.tag == svgns('g'):
            recurse_tree(paths, tree_root, node, matrix_new, parent_visibility=v)
        elif node.tag == svgns('use'):
            referenced_name = node.attrib[xlinkns("href")]
            m = re.search(r"#(\S+)", referenced_name)
            if m:
                el_id = m.groups()[0]
                xpath = ".//*[@id='%s']" % (el_id)
                node_to_copy = tree_root.find(xpath)
                # Modify the matrix if x and y are specified
                if "x" in node.attrib and "y" in node.attrib:
                    x_diff = float(node.attrib["x"])
                    y_diff = float(node.attrib["y"])
                    shift_matrix = [[1.0, 0.0, x_diff],
                                    [0.0, 1.0, y_diff]]
                    matrix_new = transform.compose(matrix_new, shift_matrix)
                # Need to put node_to_copy in a list here so  that there is
                # something to iterate over in the call to recurse_tree
                recurse_tree(
                    paths,
                    tree_root,
                    [node_to_copy],
                    matrix_new,
                    parent_visibility=v)
            else:
                raise RuntimeError(
                    "Unrecognized specifier for href ('%s'). "
                    "Note that axibot only supports css id's in use tags" % referenced_name)
        elif node.tag == svgns('path'):
            paths.append(transform_path(node.get('d'), matrix_new))
        elif node.tag in (svgns('rect'), svgns('line'), svgns('polyline'),
                          svgns('polygon'), svgns('ellipse'),
                          svgns('circle')):
            s = convert_to_path(node, matrix_new)
            paths.append(transform_path(s, matrix_new))
        elif node.tag == svgns('text'):
            log.warn("Cannot directly draw text. Convert text to path.")
        elif node.tag == svgns('image'):
            log.warn("Cannot draw raster images. Vectorize first.")
        else:
            log.debug("Ignoring <%s> tag.", node.tag)


def extract_paths(s):
    root = ElementTree.fromstring(s)
    # Check if the svg element has the correct namespace
    if not root.tag.startswith('{http://www.w3.org/2000/svg}'):
        log.warn("File is invalid, missing svg namespace declaration")
        raise RuntimeError(
            "File is invalid, missing svg namespace declaration")

    svg_width, svg_height = get_document_dimensions(root)
    viewbox = root.get('viewBox')

    if viewbox:
        info = viewbox.strip().replace(',', ' ').split(' ')
        sx = svg_width / float(info[2])
        sy = svg_height / float(info[3])
        tx = -float(info[0])
        ty = -float(info[1])
    else:
        sx = sy = 1
        tx = ty = 0

    matrix = transform.parse(
        'scale(%f,%f) translate(%f,%f)' %
        (sx, sy, tx, ty))

    paths = []
    recurse_tree(paths, root, root, matrix)
    return paths


def split_disconnected_paths(paths):
    """
    Accepts a list of Path instances. Iterates over paths to determine if
    adjacent sections are actually connected. If they are not, the paths are
    split into multiple Path instances so that each instance contains only
    connected paths.
    """
    out_paths = []
    for path in paths:
        new_path = Path()
        last_point = path[0].start
        for section in path:
            if section.start != last_point:
                out_paths.append(new_path)
                new_path = Path()
            last_point = section.end
            new_path.append(section)
        out_paths.append(new_path)
    return out_paths


def distance_squared(a, b):
    return ((b.real - a.real)**2) + ((b.imag - a.imag)**2)


def find_closest_path(current_point, paths):
    best_score = None
    best_path = None
    for path in paths:
        score = distance_squared(current_point, path[0].start)
        if (not best_score) or (score < best_score):
            best_score = score
            best_path = path
    return best_path


def sort_paths(paths):
    """
    Sort list of paths by start point. This is a crude heuristic to try to
    avoid spending as much time moving around with the pen up.
    """
    out_paths = []
    current_point = complex(0, 0)
    while paths:
        next_path = find_closest_path(current_point, paths)
        current_point = next_path[-1].end
        paths.remove(next_path)
        out_paths.append(next_path)
    return out_paths


def preprocess_paths(paths):
    paths = split_disconnected_paths(paths)
    paths = sort_paths(paths)
    return paths



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import logging

import sys
import argparse
import math

import matplotlib.pyplot as plt

from axibot import svg, planning, config
from axibot.action import PenUpMove, PenDownMove, XYMove


def show(opts):
    if opts.out:
        plt.savefig(opts.out, dpi=300)
    else:
        plt.show()


def debug_paths(opts):
    """
    Render an SVG file into paths, and then plot them with matplotlib.
    """
    subdivide = 100
    with open(opts.filename) as f:
        svg_str = f.read()
    paths = svg.extract_paths(svg_str)
    paths = svg.preprocess_paths(paths)
    for path in paths:
        xdata = []
        ydata = []
        for n in range(subdivide + 1):
            point = path.point(n / subdivide)
            xdata.append(point.real)
            ydata.append(-point.imag)
        plt.plot(xdata, ydata, 'g-')

    show(opts)


def debug_segments(opts):
    """
    Render an SVG file into linear segments, and then plot them with
    matplotlib.
    """
    with open(opts.filename) as f:
        svg_str = f.read()
    paths = svg.extract_paths(svg_str)
    paths = svg.preprocess_paths(paths)
    segments = svg.plan_segments(paths, resolution=config.CURVE_RESOLUTION)

    xdata = []
    ydata = []

    for segment in segments:
        for (x, y) in segment:
            xdata.append(x)
            ydata.append(-y)

    plt.plot(xdata, ydata, 'g-')
    show(opts)


def debug_connected_segments(opts):
    with open(opts.filename) as f:
        svg_str = f.read()
    paths = svg.extract_paths(svg_str)
    paths = svg.preprocess_paths(paths)
    segments = svg.plan_segments(paths, resolution=config.CURVE_RESOLUTION)
    segments = svg.add_pen_up_moves(segments)

    for segment, pen_up in segments:
        xdata = []
        ydata = []
        for (x, y) in segment:
            xdata.append(x)
            ydata.append(-y)
        plt.plot(xdata, ydata, 'r-' if pen_up else 'g-')

    show(opts)


def debug_corners(opts):
    with open(opts.filename) as f:
        svg_str = f.read()
    paths = svg.extract_paths(svg_str)
    paths = svg.preprocess_paths(paths)
    segments = svg.plan_segments(paths, resolution=config.CURVE_RESOLUTION)
    segments = svg.add_pen_up_moves(segments)
    step_segments = planning.convert_inches_to_steps(segments)
    segments_limits = planning.plan_speed(step_segments)

    up_xdata = []
    up_ydata = []
    up_speed = []
    down_xdata = []
    down_ydata = []
    down_speed = []

    def speed_to_opacity(speed, limit, min_opacity=0.2):
        opacity = speed / limit
        if opacity < min_opacity:
            return min_opacity
        elif opacity > 1.0:
            return 1.0
        else:
            return opacity

    def speed_to_color(speed, pen_up):
        if pen_up:
            return (1.0, 0.0, 0.0,
                    speed_to_opacity(speed, config.SPEED_PEN_UP))
        else:
            return (0.0, 0.1, 0.0,
                    speed_to_opacity(speed, config.SPEED_PEN_DOWN))

    def record_point(point, vmax, pen_up):
        x, y = point
        y = -y
        if pen_up:
            up_xdata.append(x)
            up_ydata.append(y)
            up_speed.append(speed_to_color(vmax, pen_up))
        else:
            down_xdata.append(x)
            down_ydata.append(y)
            down_speed.append(speed_to_color(vmax, pen_up))

    for segment, pen_up in segments_limits:
        xdata = []
        ydata = []
        for point, vmax in segment:
            record_point(point, vmax, pen_up)
            xdata.append(point[0])
            ydata.append(-point[1])
        plt.plot(xdata, ydata, 'r-' if pen_up else 'g-')

    plt.scatter(up_xdata, up_ydata, s=50, linewidths=0, c=up_speed)
    plt.scatter(down_xdata, down_ydata, s=50, linewidths=0, c=down_speed)

    show(opts)


def generate_actions(opts):
    with open(opts.filename) as f:
        svg_str = f.read()
    paths = svg.extract_paths(svg_str)
    paths = svg.preprocess_paths(paths)
    segments = svg.plan_segments(paths, resolution=config.CURVE_RESOLUTION)
    segments = svg.add_pen_up_moves(segments)
    step_segments = planning.convert_inches_to_steps(segments)
    segments_limits = planning.plan_speed(step_segments)
    return planning.plan_actions(segments_limits, 1000, 1000)


def debug_actions(opts):
    actions = generate_actions(opts)

    x = y = 0
    pen_up = True

    xdata = []
    ydata = []

    for action in actions:
        if isinstance(action, PenUpMove):
            print("pen up")
            if not pen_up:
                # plot the pen down stuff
                plt.plot(xdata, ydata, 'g-')
                xdata = [x]
                ydata = [-y]
            pen_up = True
        elif isinstance(action, PenDownMove):
            print("pen down")
            if pen_up:
                # plot the pen up stuff
                plt.plot(xdata, ydata, 'r-')
                xdata = [x]
                ydata = [-y]
            pen_up = False
        elif isinstance(action, XYMove):
            dx = action.m1 + action.m2
            dy = action.m1 - action.m2
            print("%s move %d, %d" % ('up' if pen_up else 'down', dx, dy))
            x += dx
            y += dy

            xdata.append(x)
            ydata.append(-y)
        else:
            raise ValueError("Not expecting %r" % action)

    plt.plot(xdata, ydata, 'r-' if pen_up else 'g-')

    show(opts)


def debug_speed(opts):
    actions = generate_actions(opts)

    x = y = t = 0
    pen_up = True

    up_tdata = []
    up_vdata = []
    down_tdata = []
    down_vdata = []

    for action in actions:
        if isinstance(action, PenUpMove):
            pen_up = True
        elif isinstance(action, PenDownMove):
            pen_up = False
        elif isinstance(action, XYMove):
            dx = action.m1 + action.m2
            dy = action.m1 - action.m2
            x += dx
            y += dy
            t += action.duration

            dist = math.sqrt(dx**2 + dy**2)
            # XXX We have to ensure that duration is never zero -- this is
            # broken
            v = dist / max(action.duration, 0.01)
            if pen_up:
                up_tdata.append(t)
                up_vdata.append(v)
            else:
                down_tdata.append(t)
                down_vdata.append(v)
        else:
            raise ValueError("Not expecting %r" % action)

        plt.plot(up_tdata, up_vdata, 'r-')
        plt.plot(down_tdata, down_vdata, 'g-')

    show(opts)


def main(argv=sys.argv):
    p = argparse.ArgumentParser(description='Debug axibot software internals.')
    p.add_argument('--verbose', action='store_true',
                   help='Print verbose debug output.')
    p.set_defaults(function=None)

    subparsers = p.add_subparsers(help='sub-command help')

    p_paths = subparsers.add_parser(
        'paths', help='Render normalized paths.')
    p_paths.add_argument('filename')
    p_paths.add_argument('--out', help='Save rendering to file.')
    p_paths.set_defaults(function=debug_paths)

    p_segments = subparsers.add_parser(
        'segments', help='Render linear segments.')
    p_segments.add_argument('filename')
    p_segments.add_argument('--out', help='Save rendering to file.')
    p_segments.set_defaults(function=debug_segments)

    p_connected_segments = subparsers.add_parser(
        'connected', help='Render linear segments with pen-up moves.')
    p_connected_segments.add_argument('filename')
    p_connected_segments.add_argument('--out', help='Save rendering to file.')
    p_connected_segments.set_defaults(function=debug_connected_segments)

    p_corners = subparsers.add_parser(
        'corners', help='Render points with speeds.')
    p_corners.add_argument('filename')
    p_corners.add_argument('--out', help='Save rendering to file.')
    p_corners.set_defaults(function=debug_corners)

    p_actions = subparsers.add_parser(
        'actions', help='Render final computed actions.')
    p_actions.add_argument('filename')
    p_actions.add_argument('--out', help='Save rendering to file.')
    p_actions.set_defaults(function=debug_actions)

    p_speed = subparsers.add_parser(
        'speed', help='Render final computed speed profile.')
    p_speed.add_argument('filename')
    p_speed.add_argument('--out', help='Save rendering to file.')
    p_speed.set_defaults(function=debug_speed)

    opts, args = p.parse_known_args(argv[1:])

    logging.basicConfig(level=logging.DEBUG if opts.verbose else logging.INFO)

    if opts.function:
        return opts.function(opts)
    else:
        p.print_help()



"""
Communication between web user and the axibot server.

Server to client messages:

    state update
    new document
    error
    completed job

Client to server messages:

    set document
    manual pen up
    manual pen down
    resume plotting
    cancel plotting
"""
import json


class Message:
    def serialize(self):
        d = self.__dict__.copy()
        d['type'] = self.reverse_types[self.__class__]
        return json.dumps(d)

    @classmethod
    def deserialize(cls, raw):
        obj = json.loads(raw)
        msg_class = cls.types[obj.pop('type')]
        return msg_class(**obj)


class StateMessage(Message):
    """
    Inform a connected client that the server's state has changed.
    """
    def __init__(self, state, num_actions, action_index, x, y, pen_up,
                 estimated_time, consumed_time):
        self.state = state
        self.num_actions = num_actions
        self.action_index = action_index
        self.x = x
        self.y = y
        self.pen_up = pen_up
        self.estimated_time = estimated_time
        self.consumed_time = consumed_time


class NewDocumentMessage(Message):
    """
    Feed a new active document to a connected client.
    """
    def __init__(self, filename, document):
        self.filename = filename
        self.document = document


class ErrorMessage(Message):
    """
    Inform a connected client of an error on the server.
    """
    def __init__(self, text):
        self.text = text


class CompletedJobMessage(Message):
    """
    Inform a connected client that the current plotting job has finished, and
    provide stats about it.
    """
    def __init__(self, estimated_time, actual_time):
        self.estimated_time = estimated_time
        self.actual_time = actual_time


class SetDocumentMessage(Message):
    """
    Instruct the server to set a new active document.
    """
    def __init__(self, filename, document):
        self.filename = filename
        self.document = document


class ManualPenUpMessage(Message):
    """
    Instruct the server to lift the pen up.
    """
    pass


class ManualPenDownMessage(Message):
    """
    Instruct the server to drop the pen down.
    """
    pass


class ResumePlottingMessage(Message):
    """
    Instruct the server to resume plotting.
    """
    pass


class CancelPlottingMessage(Message):
    """
    Instruct the server to cancel plotting the current job.
    """
    pass


Message.types = {
    'state': StateMessage,
    'new-document': NewDocumentMessage,
    'error': ErrorMessage,
    'completed-job': CompletedJobMessage,
    'set-document': SetDocumentMessage,
    'manual-pen-up': ManualPenUpMessage,
    'manual-pen-down': ManualPenDownMessage,
    'resume-plotting': ResumePlottingMessage,
    'cancel-plotting': CancelPlottingMessage,
}

Message.reverse_types = {v: k for k, v in Message.types.items()}



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import logging

import time
import sys
import argparse
import os.path
from datetime import timedelta

from six.moves import input

try:
    import coloredlogs
except ImportError:
    coloredlogs = None

from . import planning, config
from .ebb import EiBotBoard, MockEiBotBoard
from .job import Job

log = logging.getLogger(__name__)


def manual_command(bot, cmd):
    args = cmd.split()
    method = args[0]
    arg = args[1:]
    arg = tuple(map(int, arg))
    try:
        getattr(bot, method)(*arg)
    except AttributeError as e:
        log.error("Command not found: %s", method)
    except (TypeError, ValueError) as e:
        log.error("Error: %s", e)


def manual(opts):
    if opts.mock:
        bot = MockEiBotBoard()
    else:
        bot = EiBotBoard.find()
    try:
        bot.servo_setup(config.PEN_DOWN_POSITION, config.PEN_UP_POSITION,
                        config.SERVO_SPEED, config.SERVO_SPEED)
        if opts.cmd:
            cmd = ' '.join(opts.cmd)
            manual_command(bot, cmd)
        else:
            while True:
                cmd = input('(axibot) ')
                manual_command(bot, cmd)
    finally:
        bot.close()


def human_friendly_timedelta(td):
    days = td.days
    hours, remainder = divmod(td.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    pieces = []
    if days:
        pieces.append(("%d days" % days) if days > 1 else "1 day")
    if hours:
        pieces.append(("%d hours" % hours) if hours > 1 else "1 hour")
    if minutes:
        pieces.append(("%d minutes" % minutes) if minutes > 1 else "1 minute")
    if seconds:
        pieces.append(("%d seconds" % seconds) if seconds > 1 else "1 second")
    return ", ".join(pieces)


def load_job(filename):
    if filename.endswith('.svg'):
        with open(filename) as f:
            return planning.plan_job(f.read(), filename)
    elif filename.endswith('.axibot.json'):
        with open(filename) as f:
            return Job.deserialize(f)
    else:
        print("Only .svg and .axibot.json files are supported!")
        raise SystemExit


def plan(opts):
    outfile = opts.out
    if not outfile:
        outfile = opts.filename.rsplit('.', 1)[0] + '.axibot.json'
    if os.path.exists(outfile) and not opts.overwrite:
        print("File '%s' exists, pass --overwrite or --out." % outfile)
    else:
        job = load_job(opts.filename)
        with open(outfile, 'w') as f:
            job.serialize(f)


def info(opts):
    job = load_job(opts.filename)
    td = job.duration()
    log.info("Number of moves: %s", len(job))
    log.info("Expected time: %s", human_friendly_timedelta(td))


def plot(opts):
    job = load_job(opts.filename)
    count = len(job)
    log.info("Loaded %d actions.", count)

    pen_up_delay, pen_down_delay = \
        planning.calculate_pen_delays(job.pen_up_position,
                                      job.pen_down_position,
                                      job.servo_speed)

    if opts.mock:
        bot = MockEiBotBoard()
    else:
        bot = EiBotBoard.find()
    try:
        # The initial pen up might need a fairly slow delay, because we don't
        # know where the servo was before.
        bot.pen_up(pen_up_delay)
        log.info("Configuring servos.")
        bot.disable_motors()
        bot.servo_setup(config.PEN_DOWN_POSITION, config.PEN_UP_POSITION,
                        config.SERVO_SPEED, config.SERVO_SPEED)
        log.info("Pen up and motors off. Move carriage to top left corner.")
        input("Press enter to begin.")

        start_time = time.time()
        bot.enable_motors(1)

        for ii, move in enumerate(job):
            log.info("Move %d/%d: %s" % (ii, count, move))
            bot.do(move)

        bot.pen_up(pen_up_delay)
        end_time = time.time()
        estimated_td = job.duration()
        actual_td = timedelta(seconds=(end_time - start_time))
        print("Finished!")
        print("Expected time: %s" % human_friendly_timedelta(estimated_td))
        print("Actual time: %s" % human_friendly_timedelta(actual_td))
        if opts.mock:
            print("---")
            print("Mock EiBotBoard recorded:")
            print("Max speed: %0.3f steps/ms" % bot.max_speed)
            print("Max acceleration: %0.3f steps/ms delta" %
                  bot.max_acceleration)
    finally:
        bot.close()


def server(opts):
    from axibot.server import serve
    serve(opts)


def main(args=sys.argv):
    p = argparse.ArgumentParser(description='Print with the AxiDraw.')
    p.add_argument('--verbose', action='store_true')
    p.add_argument('--mock', action='store_true')
    p.set_defaults(function=None)

    subparsers = p.add_subparsers(help='sub-command help')

    p_plan = subparsers.add_parser(
        'plan', help='Plan the plotting actions for an SVG file.')
    p_plan.add_argument('filename')
    p_plan.add_argument('--out')
    p_plan.add_argument('--overwrite', action='store_true')
    p_plan.set_defaults(function=plan)

    p_plot = subparsers.add_parser(
        'plot', help='Plot an SVG file directly.')
    p_plot.add_argument('filename')
    p_plot.set_defaults(function=plot)

    p_info = subparsers.add_parser(
        'info', help='Print information about an SVG file.')
    p_info.add_argument('filename')
    p_info.set_defaults(function=info)

    p_server = subparsers.add_parser(
        'server', help='Run a server for remote plotting.')
    p_server.add_argument('--port', type=int, default=8888)
    p_server.set_defaults(function=server)

    p_manual = subparsers.add_parser(
        'manual', help='Manual control shell.')
    p_manual.add_argument('cmd', nargs='*')
    p_manual.set_defaults(function=manual)

    opts, args = p.parse_known_args(args[1:])

    if coloredlogs:
        coloredlogs.install(level='DEBUG' if opts.verbose else 'INFO')
    else:
        logging.basicConfig(level=logging.DEBUG if opts.verbose else
                            logging.INFO)

    if opts.function:
        return opts.function(opts)
    else:
        p.print_help()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import math
from operator import itemgetter

from colormath.color_objects import sRGBColor, LabColor
from colormath.color_conversions import convert_color


pen_sets = {
    'precise-v5': {
        'black': (59, 59, 59),
        'blue': (61, 93, 134),
        'red': (138, 56, 60),
        'green': (52, 126, 101),
        'purple': (93, 90, 179),
        'lightblue': (69, 153, 189),
        'pink': (225, 87, 146),
    }
}


def rgb_to_lab(rgb):
    rgb_color = sRGBColor(rgb[0], rgb[1], rgb[2])
    lab_color = convert_color(rgb_color, LabColor)
    return lab_color.lab_l, lab_color.lab_a, lab_color.lab_b


def perceptual_distance(a, b):
    a = rgb_to_lab(a)
    b = rgb_to_lab(b)
    return math.sqrt((b[2] - a[2])**2 +
                     (b[1] - a[1])**2 +
                     (b[0] - a[0])**2)


def find_pen_match(color, pen_set):
    scores = {}
    for pen, pen_color in pen_sets[pen_set].items():
        scores[pen] = perceptual_distance(color, pen_color)
    scores = scores.items()
    scores.sort(key=itemgetter(1))
    return scores[0]



from __future__ import print_function

from setuptools import setup, find_packages


setup(name='axibot',
      version='0.0.3',
      description='Software for AxiDraw pen plotting robot',
      long_description='',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Environment :: Console',
          'Environment :: Web Environment',
          'Topic :: Artistic Software',
          'Topic :: Multimedia :: Graphics',
          'Topic :: Printing',
      ],
      keywords='eibotboard axidraw cam motion pen drawing plotter',
      url='https://github.com/storborg/axibot',
      author='Scott Torborg',
      author_email='storborg@gmail.com',
      license='GPL',
      packages=find_packages(),
      install_requires=[
          'six',
          'pyserial',
          'svg.path<3.0',
          'colormath',
      ],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      extras_require={
          'server': ['aiohttp',
                     'aiohttp_mako'],
          'debug': ['matplotlib'],
      },
      include_package_data=True,
      zip_safe=False,
      entry_points="""\
      [console_scripts]
      axibot = axibot.cmd:main
      axibot-debug = axibot.debug:main [debug]
      """)



import logging
import os
import os.path

from aiohttp import web
import aiohttp_mako

from ..ebb import EiBotBoard, MockEiBotBoard
from .. import planning, config

from . import views, handlers, plotting
from .state import State

log = logging.getLogger(__name__)


__here__ = os.path.dirname(os.path.realpath(__file__))
template_dir = os.path.join(__here__, 'templates')
static_dir = os.path.join(__here__, 'dist')

base_dir = os.path.dirname(os.path.dirname(__here__))
examples_dir = os.path.join(base_dir, 'examples')


def make_app(bot):
    app = web.Application()

    app['state'] = State.idle
    app['action_index'] = 0
    app['clients'] = set()
    app['bot'] = bot
    app['position'] = 0, 0
    app['pen_up'] = None

    bot.enable_motors(1)
    bot.servo_setup(config.PEN_DOWN_POSITION, config.PEN_UP_POSITION,
                    config.SERVO_SPEED, config.SERVO_SPEED)

    app['pen_up_delay'], app['pen_down_delay'] = \
        planning.calculate_pen_delays(config.PEN_UP_POSITION,
                                      config.PEN_DOWN_POSITION,
                                      config.SERVO_SPEED)

    # This will initialize the server state.
    filename = 'line.svg'
    with open(os.path.join(examples_dir, filename)) as f:
        doc = f.read()
        app['job'] = job = plotting.process_upload(app, doc, filename)
        app['estimated_time'] = job.duration().total_seconds()
        app['consumed_time'] = 0

    aiohttp_mako.setup(app,
                       directories=[template_dir],
                       input_encoding='utf-8',
                       output_encoding='utf-8',
                       imports=['from markupsafe import escape'],
                       default_filters=['escape'])

    app.router.add_route('GET', '/', views.index)
    app.router.add_route('GET', '/api', handlers.client_handler)

    app.router.add_static('/static', static_dir)

    return app


def serve(opts):
    if opts.mock:
        bot = MockEiBotBoard()
    else:
        bot = EiBotBoard.find()

    try:
        app = make_app(bot)
        web.run_app(app, port=opts.port)
    finally:
        bot.close()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import logging
import time
import math

import serial
from serial.tools.list_ports import comports

from . import config

log = logging.getLogger(__name__)


class EiBotException(Exception):
    pass


class EiBotBase:
    def do(self, move):
        kw = move.__dict__.copy()
        name = move.name
        if name in ('pen_up', 'pen_down',
                    'xy_accel_move', 'xy_move',
                    'ab_move'):
            method = getattr(self, name)
            return method(**kw)
        else:
            raise EiBotException("Don't know how to do move %r / %s" %
                                 (move, move))


class EiBotBoard(EiBotBase):
    def __init__(self, ser):
        self.serial = ser

    @classmethod
    def list_ports(cls):
        ports = comports()
        for port in ports:
            if port[1].startswith('EiBotBoard'):
                yield port[0]
            elif port[2].upper().startswith('USB VID:PID=04D8:FD92'):
                yield port[0]

    @classmethod
    def open(cls, port):
        ser = serial.Serial(port, timeout=1.0)
        # May need to try several times to get a response from the board?
        # This behavior is taken from the ebb_serial usage, not sure if it's
        # necessary.
        for attempt in range(3):
            ser.write(b'v\r')
            version = ser.readline()
            if version and version.startswith(b'EBB'):
                return cls(ser)

    @classmethod
    def find(cls):
        for port in cls.list_ports():
            if port:
                return cls.open(port)
        raise EiBotException("Could not find a connected EiBotBoard.")

    def close(self):
        # XXX Maybe switch to a context manger for this?
        self.serial.close()

    def robust_readline(self):
        for attempt in range(config.MAX_RETRIES):
            resp = self.serial.readline()
            if resp:
                return resp

    def query(self, cmd):
        self.serial.write(cmd.encode('ascii'))
        resp = self.robust_readline()
        if cmd.strip().lower() not in ('v', 'i', 'a', 'mr', 'pi', 'qm'):
            # Discard response.
            self.robust_readline()
        return resp

    def command(self, cmd):
        cmd = cmd.encode('ascii')
        log.debug("Sending command: %s", cmd)
        self.serial.write(cmd)
        resp = self.robust_readline()
        if not resp.strip().startswith(b'OK'):
            if resp:
                raise EiBotException(
                    "Unexpected response from EBB:\n"
                    "Command: %s\n"
                    "Response: %s" % (cmd.strip(), resp.strip()))

    def enable_motors(self, res):
        """
        Enable motors. Available resolutions:
            0, -> Motor disabled
            1, -> 16X microstepping
            2, -> 8X microstepping
            3, -> 4X microstepping
            4, -> 2X microstepping
            5, -> No microstepping
        """
        if res < 0:
            res = 0
        elif res > 5:
            res = 5
        self.command('EM,%s,%s\r' % (res, res))

    def disable_motors(self):
        self.command('EM,0,0\r')

    def query_prg_button(self):
        self.command('QB\r')

    def toggle_pen(self):
        self.command('TP\r')

    def servo_setup(self,
                    pen_down_position, pen_up_position,
                    servo_up_speed, servo_down_speed):
        """
        Configure EBB servo control offsets and speeds.

        From the axidraw.py comments:

            "Pen position units range from 0% to 100%, which correspond to a
            typical timing range of 7500 - 25000 in units of 1/(12 MHz). 1%
            corresponds to ~14.6 us, or 175 units of 1/(12 MHz)."

            "Servo speed units are in units of %/second, referring to the
            percentages above.  The EBB takes speeds in units of 1/(12 MHz)
            steps per 24 ms.  Scaling as above, 1% of range in 1 second with
            SERVO_MAX = 28000  and  SERVO_MIN = 7500 corresponds to 205 steps
            change in 1 s That gives 0.205 steps/ms, or 4.92 steps / 24 ms
            Rounding this to 5 steps/24 ms is sufficient."
        """
        slope = float(config.SERVO_MAX - config.SERVO_MIN) / 100.
        up_setting = int(round(config.SERVO_MIN + (slope * pen_up_position)))
        down_setting = int(round(config.SERVO_MIN +
                                 (slope * pen_down_position)))
        up_speed = 5 * servo_up_speed
        down_speed = 5 * servo_down_speed
        self.command('SC,4,%s\r' % up_setting)
        self.command('SC,5,%s\r' % down_setting)
        self.command('SC,11,%s\r' % up_speed)
        self.command('SC,12,%s\r' % down_speed)

    def pen_up(self, delay):
        self.command('SP,1,%s\r' % delay)

    def pen_down(self, delay):
        self.command('SP,0,%s\r' % delay)

    def xy_accel_move(self, dx, dy, v_initial, v_final):
        """
        Move X/Y axes as: "AM,<v_initial>,<v_final>,<axis1>,<axis2><CR>"
        Typically, this is wired up such that axis 1 is the Y axis and axis 2
        is the X axis of motion. On EggBot, Axis 1 is the "pen" motor, and Axis
        2 is the "egg" motor. Note that minimum move duration is 5 ms.
        Important: Requires firmware version 2.4 or higher.

        Not used in "stock" AxiDraw Inkscape driver: theoretically this could
        replace a substantial portion of the motion planning, and eliminate the
        plot_segment... planning, but there are some comments in firmware code
        and such that indicate that it doesn't work correctly yet.
        """
        self.command('AM,%s,%s,%s,%s\r' % (v_initial, v_final, dx, dy))

    def xy_move(self, m1, m2, duration):
        """
        Move M1/M2 axes as: "SM,<move_duration>,<axis1>,<axis2><CR>"

        Move duration is in milliseconds and can be 1 to 16777215.
        m1 and m2 are in steps and are signed 24-bit integers.

        On the AxiDraw, these axes are blended: the coordinate space of the
        control API is rotated 45 degrees from the coordinate space of the
        actual movement. It's unclear why this is.
        """
        assert isinstance(m1, int)
        assert isinstance(m2, int)
        assert isinstance(duration, int)
        assert (duration >= 1) and (duration <= 16777215), \
            "Invalid duration %r" % duration

        # XXX add checks for minimum or maximum step rates

        self.command('SM,%s,%s,%s\r' % (duration, m1, m2))

    def ab_move(self, da, db, duration):
        """
        Issue command to move A/B axes
        as: "XM,<move_duration>,<axisA>,<axisB><CR>"
        Then, <Axis1> moves by <AxisA> + <AxisB>,
        and <Axis2> as <AxisA> - <AxisB>

        Not used in "stock" AxiDraw Inkscape driver.
        """
        self.command('XM,%s,%s,%s\r' % (duration, da, db))


class MockEiBotBoard(EiBotBase):
    def __init__(self):
        self.x = 0
        self.y = 0
        self.last_speed = 0
        self.max_speed = 0
        self.max_acceleration = 0

    def close(self):
        pass

    def query(self, cmd):
        raise NotImplementedError(
            "Mock EBB doesn't know how to handle queries.")

    def command(self, cmd):
        cmd = cmd.encode('ascii')
        log.debug("Sending command: %s", cmd)
        self.serial.write(cmd)
        resp = self.robust_readline()
        if not resp.strip().startswith(b'OK'):
            if resp:
                raise EiBotException(
                    "Unexpected response from EBB:\n"
                    "Command: %s\n"
                    "Response: %s" % (cmd.strip(), resp.strip()))

    def enable_motors(self, res):
        log.warn("Mock EBB: enable_motors")

    def disable_motors(self):
        log.warn("Mock EBB: disable_motors")

    def query_prg_button(self):
        log.warn("Mock EBB: query_prg_button")

    def toggle_pen(self):
        log.warn("Mock EBB: toggle_pen")

    def servo_setup(self,
                    pen_down_position, pen_up_position,
                    servo_up_speed, servo_down_speed):
        log.warn("Mock EBB: servo_setup")

    def pen_up(self, delay):
        log.warn("Mock EBB: pen_up delay %s", delay)
        time.sleep(delay / 1000.)

    def pen_down(self, delay):
        log.warn("Mock EBB: pen_down delay:%s", delay)
        time.sleep(delay / 1000.)

    def xy_accel_move(self, dx, dy, v_initial, v_final):
        log.warn("Mock EBB: xy_accel_move / Cannot simulate delay.")

    def xy_move(self, m1, m2, duration):
        dx = m1 + m2
        dy = m1 - m2
        dist = math.sqrt((dx**2) + (dy**2))
        speed = dist / duration
        if speed > self.max_speed:
            self.max_speed = speed
        accel = speed - self.last_speed
        if accel > self.max_acceleration:
            self.max_acceleration = accel
        self.last_speed = speed
        self.x += dx
        self.y += dy
        log.warn("Mock EBB: xy_move m1:%s m2:%s duration:%s -> %s, %s",
                 m1, m2, duration, self.x, self.y)
        time.sleep(duration / 1000.)

    def ab_move(self, da, db, duration):
        log.warn("Mock EBB: ab_move")
        time.sleep(duration / 1000.)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from pprint import pformat


class Move:
    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, pformat(self.__dict__))

    def __str__(self):
        attrs = ['%s:%s' % (k, v) for k, v in self.__dict__.items()]
        return '%s %s' % (self.name, ' '.join(attrs))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class PenUpMove(Move):
    name = 'pen_up'

    def __init__(self, delay):
        self.delay = delay

    def time(self):
        return self.delay


class PenDownMove(Move):
    name = 'pen_down'

    def __init__(self, delay):
        self.delay = delay

    def time(self):
        return self.delay


class XYMove(Move):
    name = 'xy_move'

    def __init__(self, m1, m2, duration):
        assert isinstance(m1, int), "got %r, wanted an int" % m1
        assert isinstance(m2, int), "got %r, wanted an int" % m2
        assert isinstance(duration, int), "got %r, wanted an int" % duration
        assert (m1 != 0) or (m2 != 0), \
            "m1:%r, m2:%r, one must be nonzero" % (m1, m2)
        assert duration > 0, \
            "duration: %r must be greater than zero" % duration
        assert duration >= 30, \
            "duration: %r must be >= 30ms" % duration
        self.m1 = m1
        self.m2 = m2
        self.duration = duration

    def time(self):
        return self.duration


class XYAccelMove(Move):
    name = 'xy_accel_move'

    # XXX These might be better as "m1" and "m2" rather than dx/dy.
    def __init__(self, dx, dy, v_initial, v_final):
        self.dx = dx
        self.dy = dy
        self.v_initial = v_initial
        self.v_final = v_final

    def time(self):
        raise NotImplementedError


class ABMove(Move):
    name = 'ab_move'

    def __init__(self, da, db, duration):
        self.da = da
        self.db = db
        self.duration = duration

    def time(self):
        raise NotImplementedError
